/*
 * $RCSfile: main.c,v $
 * $Date: 2005/07/09 15:47:14 $ - $Revision: 1.15 $
 *
 * This file is distributed as a part of MMSRIP ( MMS Ripper ).
 * Copyright (c) 2005 Nicolas BENOIT
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <getopt.h>
#include "mms.h"
#include "error.h"
#include "common.h"

/*
 * options
 */
static char * options = "-ahvtqo:d:";

static struct option long_options[] = {
  {"about", 0, NULL, 'a'},
  {"version", 0, NULL, 'v'},
  {"help", 0, NULL, 'h'},
  {"test", 0, NULL, 't'},
  {"quiet", 0, NULL, 'q'},
  {"output", 1, NULL, 'o'},
  {"delay", 1, NULL, 'd'},
  {NULL, 0, NULL, 0}
};


/*
 * usage
 */
void
usage ( void )
{
  fprintf ( stderr, "%s (%s) version %s\n\n", PROGRAM_SHORT_NAME, PROGRAM_FULL_NAME, PROGRAM_VERSION );
  fprintf ( stderr, "usage: %s <[-oFILE] stream url> ...\n\n", PROGRAM_SHORT_NAME );
  fprintf ( stderr, "General Options:\n\t"        \
                    "-a, --about\t\tshow information about %s\n\t" \
                    "-h, --help\t\tshow this help\n\t"      \
                    "-v, --version\t\tshow version number\n\n" \
                    "Program Behaviour:\n\t"      \
                    "-oFILE, --output=FILE\toutput to specified file (can be repeated)\n\t" \
                    "-q, --quiet\t\tquiet mode (can be repeated)\n\t" \
                    "-dDELAY, --delay=DELAY\tsave the stream during DELAY seconds and exit\n\t" \
                    "-t, --test\t\ttest mode (check stream availability)\n\n", PROGRAM_SHORT_NAME );
  return;
}


/*
 * main
 */
int
main ( int argc,
       char *argv[] )
{
  MMS *mms;
  FILE *f;
  char *output_filename;
  STREAM_LIST *stream_list;
  STREAM_LIST *block;
  char c;
  int ret = 0;
  int quiet = 0;
  int test = 0;
  int delay = 0;
  time_t end = 0;
  ssize_t len_written;
  uint64_t total_len_written;

  if ( ( stream_list = (STREAM_LIST *) malloc(sizeof(STREAM_LIST)) ) == NULL )
    {
      error ( "main", "early initialization failed" );
      return -1;
    }

  stream_list->next = NULL;
  stream_list->stream = NULL;
  stream_list->output = NULL;

  block = stream_list;

  while ( ( c = getopt_long(argc, argv, options, long_options, NULL) ) != EOF )
    {
      switch (c)
        {
        case 'h':
          {
            fprintf ( stderr, "\n" );
            usage();
            return 0;
          }

        case 'v':
          {
            fprintf ( stderr, "%s version %s\n", PROGRAM_SHORT_NAME, PROGRAM_VERSION);
            return 0;
          }

        case 'a':
          {
            fprintf ( stderr, "\n" );
            fprintf ( stderr, "%s (%s) version %s\n\n", PROGRAM_SHORT_NAME, PROGRAM_FULL_NAME, PROGRAM_VERSION);
            fprintf ( stderr, "Written by %s.\n", PROGRAM_AUTHORS );
            fprintf ( stderr, "With a lot of help from %s.\n\n", PROGRAM_HELPERS );
            fprintf ( stderr, "This program is free software; you can redistribute it\nand/or modify it under the terms \
of the GNU General Public License\nas published by the Free Software Foundation; either version 2,\nor (at your option) any later version.\n\n");
            return 0;
          }

        case 't':
          {
            test = 1;
            break;
          }

        case 'q':
          {
            quiet += 1;
            break;
          }

        case 'o':
          {
            if ( optarg != NULL )
              {
                if ( block->stream != NULL )
                  {
                    if ( ( block->next = malloc ( sizeof(STREAM_LIST) ) ) == NULL )
                      {
                        error ( "main", "early initialization failed" );
                        return -1;
                      }

                    block = block->next;
                    block->next = NULL;
                    block->stream = NULL;
                  }

                if ( block->output != NULL )
                  free ( block->output );

                block->output = (char *) strdup ( optarg );
              }

            break;
          }

        case 'd':
          {
            if ( optarg != NULL )
              {
                delay = atoi( optarg );

                if ( delay < 0 )
                  delay = 0;
              }

            break;
          }

        case 1:
          {
            if ( optarg != NULL )
              {
                if ( block->stream != NULL )
                  {
                    if ( ( block->next = malloc ( sizeof(STREAM_LIST) ) ) == NULL )
                      {
                        error ( "main", "early initialization failed" );
                        return -1;
                      }

                    block = block->next;
                    block->next = NULL;
                    block->output = NULL;
                  }

                block->stream = (char *) strdup ( optarg );
              }

            break;
          }
        }
    }

  if ( stream_list->stream == NULL )
    {
      usage ( );
      return -1;
    }

  if ( !quiet )
    {
      fprintf ( stderr, "\n" );
      fprintf ( stderr, "%s (%s) version %s\n\n", PROGRAM_SHORT_NAME, PROGRAM_FULL_NAME, PROGRAM_VERSION );
    }

  for ( block=stream_list; block!=NULL; block=(STREAM_LIST*)block->next )
    {
      if ( block->stream == NULL )
        {
          if ( quiet < 2 )
            {
              if ( block->output == NULL )
                error ( "main", "invalid invocation of %s", PROGRAM_SHORT_NAME );
              else
                error ( "main", "output to \'%s\' is not attached to any stream", block->output );
            }

          return -1;
        }

      if ( block->output == NULL )
        {
          output_filename = strrchr ( block->stream, '/' );

          if ( output_filename == NULL )
            {
              if ( quiet < 2 )
                error ( "main", "malformed url: \'%s\'", block->stream );

              ret = -1;
              continue;
            }

          ++output_filename;

          if ( strlen ( output_filename ) == 0 )
            {
              if ( quiet < 2 )
                error ( "main", "malformed url: \'%s\'", block->stream );

              ret = -1;
              continue;
            }

          block->output = (char *) strdup ( output_filename );
        }
    }

  if ( ret != 0 )
    goto clean;

  if ( delay != 0 )
    end = time(NULL) + delay;

  for ( block=stream_list; block!=NULL; block=(STREAM_LIST*)block->next )
    {
      output_filename = block->output;

      if ( !test )
        {
          if ( ( f = fopen ( output_filename, "w" ) ) == NULL )
            {
              if ( quiet < 2 )
                error ( "main", "unable to write in \'%s\'", output_filename );

              ret = -1;
              continue;
            }
        }
      else
        f = NULL;

      if ( ( mms = mms_create ( block->stream, f, test?1:(quiet>>1) ) ) == NULL )
        {
          if ( quiet < 2 )
            error ( "main", "unable to create mms struct" );

          if ( !test )
            {
              fclose ( f );
              remove ( output_filename );
            }

          ret = -1;
          continue;
        }

      if ( !quiet )
        fprintf ( stderr, "Connecting ...\r" );

      if ( mms_connect ( mms ) != 0 )
        {
          if ( quiet < 2 )
            error ( "main", "unable to connect" );

          mms_destroy ( mms );

          if ( !test )
            {
              fclose ( f );
              remove ( output_filename );
            }

          ret = -1;
          continue;
        }

      if ( !quiet )
        fprintf ( stderr, "Handshaking ...\r" );

      if ( mms_handshake ( mms ) != 0 )
        {
          if ( ( quiet < 2 ) && ( !test ) )
            error ( "main", "unable to handshake" );

          mms_disconnect ( mms );
          mms_destroy ( mms );

          if ( !test )
            {
              fclose ( f );
              remove ( output_filename );
            }

          if ( !quiet )
            {
              if ( !test )
                fprintf ( stderr, "Stream \'%s\' is not OK.\n\n", block->stream );
              else
                fprintf ( stderr, "Stream \'%s\' is not OK.\n", block->stream );
            }

          ret = -1;
          continue;
        }

      if ( test )
        {
          if ( !quiet )
            fprintf ( stderr, "Stream \'%s\' is OK.\n", block->stream );

          mms_disconnect ( mms );
          mms_destroy ( mms );
          continue;
        }

      if ( !quiet )
        fprintf ( stderr, "Getting header ...\r" );

      if ( ( len_written = mms_write_stream_header ( mms ) ) == -1 )
        {
          if ( quiet < 2 )
            error ( "main", "unable to write stream header" );

          mms_disconnect ( mms );
          mms_destroy ( mms );
          fclose ( f );
          remove ( output_filename );
          ret = -1;
          continue;
        }

      total_len_written = len_written;

      if ( !quiet )
        fprintf ( stderr, "Rip is about to start ...\r" );

      if ( mms_begin_rip ( mms ) != 0 )
        {
          if ( quiet < 2 )
            error ( "main", "unable to begin the rip" );

          mms_disconnect ( mms );
          mms_destroy ( mms );
          fclose ( f );
          remove ( output_filename );
          ret = -1;
          continue;
        }

      if ( mms->is_live == MMS_LIVE )
        {
          if ( !quiet )
            {
              fprintf ( stderr, "                                                 \r" );
              fprintf ( stderr, "%s: %d bytes written\r", output_filename, (ssize_t)total_len_written );
            }

          while ( 1 )
            {
              len_written = mms_write_stream_data ( mms );

              if ( len_written == 0 )
                break;
              else if ( len_written == -1 )
                {
                  if ( quiet < 2 )
                    error ( "main", "unable to write stream data" );

                  break;
                }

              total_len_written += len_written;

              if ( !quiet )
                fprintf ( stderr, "%s: %d bytes written\r", output_filename, (ssize_t)total_len_written );

              fflush ( f );

              if ( delay != 0 )
                {
                  if ( end <= time(NULL) )
                    {
                      delay = -1;
                      break;
                    }
                }
            }

          /* for a live, we should rewrite the header */

          if ( ( len_written != -1 ) && ( !quiet ) )
            fprintf ( stderr, "%s: %d bytes written\n", output_filename, (ssize_t)total_len_written );
        }
      else
        {
          register const double coef = 100.0 / (double) mms->expected_file_size;

          if ( !quiet )
            {
              fprintf ( stderr, "                                                 \r" );
              fprintf ( stderr, "%s: %.2f%%\r", output_filename, (double) total_len_written * coef );
            }

          while ( 1 )
            {
              len_written = mms_write_stream_data ( mms );

              if ( len_written == 0 )
                break;
              else if ( len_written == -1 )
                {
                  if ( quiet < 2 )
                    error ( "main", "unable to write stream data" );

                  break;
                }

              total_len_written += len_written;

              if ( !quiet )
                fprintf ( stderr, "%s: %.2f%%\r", output_filename, (double) total_len_written * coef );

              fflush ( f );

              if ( delay != 0 )
                {
                  if ( end <= time(NULL) )
                    {
                      delay = -1;
                      break;
                    }
                }
            }

          if ( ( len_written != -1 ) && ( !quiet ) && ( delay != -1 ) )
            fprintf ( stderr, "%s: 100.00%%\n", output_filename );
        }

      mms_disconnect ( mms );
      mms_destroy ( mms );

      fclose ( f );

      if ( delay == -1 )
        break;
    }

 clean:
  if ( !quiet )
    fprintf ( stderr, "\n" );

  for ( block=stream_list; block!=NULL; )
    {
      STREAM_LIST *old;
      old = block;
      block = (STREAM_LIST *) block->next;

      if ( old->stream != NULL )
        free ( old->stream );

      if ( old->output != NULL )
        free ( old->output );

      free ( old );
    }

  return ret;
}
