/*
 * $RCSfile: mms.c,v $
 * $Date: 2005/07/09 17:34:35 $ - $Revision: 1.13 $
 *
 * This file is distributed as a part of MMSRIP ( MMS Ripper ).
 * Copyright (c) 2005 Nicolas BENOIT
 *
 * It is highly based on the work of SDP Multimedia and Major MMS.
 * They deserve all the credits for it.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#define _GNU_SOURCE

#ifndef _WIN32
#include <unistd.h>
#endif
#include <stdio.h>
#ifdef _WIN32
#include <winsock2.h>
#include <string.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <strings.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>

#include "mms.h"
#include "error.h"
#include "common.h"


/*
 * Put 32 bits inside a command buffer.
 */
static void
mms_put_32 ( MMS_PACKET *pak,
             uint32_t value )
{
  pak->buf[pak->num_bytes  ] = value % 256;
  value = value >> 8;
  pak->buf[pak->num_bytes+1] = value % 256 ;
  value = value >> 8;
  pak->buf[pak->num_bytes+2] = value % 256 ;
  value = value >> 8;
  pak->buf[pak->num_bytes+3] = value % 256 ;

  pak->num_bytes += 4;
  return;
}


/*
 * Returns the 32 bits located at specified offset.
 */
static uint32_t
mms_get_32 ( const unsigned char *buf,
             const int offset )
{
  uint32_t ret;

  ret = buf[offset] ;
  ret |= buf[offset+1]<<8 ;
  ret |= buf[offset+2]<<16 ;
  ret |= buf[offset+3]<<24 ;

  return ret;
}


/*
 * Converts a string into a UTF-16 string.
 */
static void
mms_string_utf16 ( uint8_t *dest,
                   unsigned char *src,
                   const ssize_t len )
{
  ssize_t i;
  memset ( dest, 0, len );

  for ( i=0; i<len; ++i )
    {
      dest[i*2] = src[i];
      dest[i*2+1] = 0;
    }

  dest[i*2] = 0;
  dest[i*2+1] = 0;
  return;
}


/*
 * Prints out a command packet (debug purpose).
 */
#ifdef ENABLE_DEBUG
static void
mms_print_packet ( const MMS_PACKET *pak,
                   const ssize_t length,
                   const int orig )
{
  ssize_t i;
  unsigned char c;

  fprintf ( stdout, "\n-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );

  if ( orig == MMS_SERVER )
    fprintf ( stdout, " command from server (%d bytes)\n", length );
  else
    fprintf ( stdout, " command from client (%d bytes)\n", length );

  fprintf ( stdout, " start sequence: 0x%08x\n", mms_get_32(pak->buf, 0) );
  fprintf ( stdout, " command id:     0x%08x\n", mms_get_32(pak->buf, 4) );
  fprintf ( stdout, " length:         0x%8x \n", mms_get_32(pak->buf, 8) );
  fprintf ( stdout, " len8:           0x%8x \n", mms_get_32(pak->buf, 16) );
  fprintf ( stdout, " sequence #:     0x%08x\n", mms_get_32(pak->buf, 20) );
  fprintf ( stdout, " len8  (II):     0x%8x \n", mms_get_32(pak->buf, 32) );
  fprintf ( stdout, " dir | comm:     0x%08x\n", mms_get_32(pak->buf, 36) );
  fprintf ( stdout, " switches:       0x%08x\n", mms_get_32(pak->buf, 40) );

  fprintf ( stdout, "\nascii contents:\n" );

  for ( i=48; i<length; i+=2 )
    {
      c = pak->buf[i];

      if ( ( c>=32 ) && ( c<=128 ) )
        fprintf ( stdout, "%c", c );
      else
        fprintf ( stdout, "." );
    }

  fprintf ( stdout, "\n\npackage hexdump:\n" );

  for ( i=0; i<length; ++i )
    {
      c = pak->buf[i];
      fprintf ( stdout, "%02x", c );

      if ( (i % 16) == 15 )
        printf ("\n");

      if ( (i % 2) == 1 )
        printf (" ");
    }

  fprintf ( stdout, "\n\n" );
  fflush ( stdout );
  return;
}
#endif


/*
 * Sends a command packet to the server.
 * Returns 0 if success, -1 either.
 */
static int
mms_send_packet ( MMS *mms,
                  const int command,
                  const uint32_t switches, 
                  const uint32_t extra,
                  const ssize_t length,
                  const uint8_t *data )
{  
  MMS_PACKET pak;
  ssize_t written_len;

  const ssize_t len8 = ( length + (length%8) ) / 8;

  pak.num_bytes = 0;

  mms_put_32 ( &pak, 0x00000001 ); /* start sequence */
  mms_put_32 ( &pak, 0xB00BFACE ); /* #-)) */
  mms_put_32 ( &pak, length + 32 );
  mms_put_32 ( &pak, 0x20534d4d ); /* protocol type "MMS " */
  mms_put_32 ( &pak, len8 + 4 );
  mms_put_32 ( &pak, mms->seq_num++ );
  mms_put_32 ( &pak, 0x0 );        /* unknown */
  mms_put_32 ( &pak, 0x0 );
  mms_put_32 ( &pak, 2+len8 );
  mms_put_32 ( &pak, 0x00030000 | command ); /* dir | command */
  mms_put_32 ( &pak, switches );
  mms_put_32 ( &pak, extra );

  memcpy ( &pak.buf[48], data, length );

  written_len = write ( mms->socket, pak.buf, length+48 );

  if ( written_len == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_send_packet", "write() said: %s", strerror(errno) );

      return -1;
    }
  else if ( written_len != (length+48) )
    {
      if ( !mms->quiet )
        error ( "mms_send_packet", "did not write enough bytes" );

      return -1;
    }

#ifdef ENABLE_DEBUG
  mms_print_packet ( &pak, length+48, MMS_CLIENT );
#endif

  return 0;
}


/*
 * Reads some bytes from the socket.
 * This is blocking until we don't have the required amount of bytes.
 * Returns 0 if success, -1 either.
 */
static int
mms_recv_packet ( const int s,
                  MMS_PACKET *pak,
                  size_t count,
                  const int quiet )
{
  ssize_t read_len, total;
  total = 0;

  if ( MMS_BUF_SIZE < count )
    {
      if ( !quiet )
        warning ( "mms_recv_packet", "caller is too greedy" );

      count = MMS_BUF_SIZE;
    }

  while ( total < count )
    {
      read_len = read ( s, &pak->buf[total], count-total );

      if ( read_len == -1 )
        {
          if ( !quiet )
            error ( "mms_recv_packet", "read() said: %s", strerror(errno) );

          return -1;
        }

      total += read_len;
    }

  return 0;
}


/*
 * Gets server's command.
 * The first 8 bytes can be skipped setting offset to 8.
 * It writes the packet length in the provided storage.
 * Returns command ID, -1 if error.
 */
static int
mms_recv_cmd_packet ( const int s,
                      MMS_PACKET *pak,
                      ssize_t *packet_len,
                      const int offset,
                      const int quiet )
{
  MMS_PACKET tmp;

  if ( ( offset > 8 ) || ( offset < 0 ) )
    {
      if ( !quiet )
        error ( "mms_recv_cmd_packet", "provided offset is invalid" );

      return -1;
    }

  if ( mms_recv_packet ( s, &tmp, 12-offset, quiet ) != 0 )
    {
      if ( !quiet )
        error ( "mms_recv_cmd_packet", "unable to get packet header" );

      return -1;
    }

  memcpy ( &pak->buf[offset], tmp.buf, 12-offset );

  if ( offset == 0 )
    {
      if ( (pak->buf[7] != 0xb0) || (pak->buf[6] != 0x0b) ||
           (pak->buf[5] != 0xfa) || (pak->buf[4] != 0xce) )
        {
          if ( !quiet )
            error ( "mms_recv_cmd_packet", "answer should have been a cmd packet" );

          return -1;
        }
    }

  memcpy ( packet_len, &pak->buf[8], 4 );
  *packet_len = mms_get_32 ( (unsigned char *) packet_len, 0 ) + 4;

  if ( ( *packet_len + 12 ) > MMS_BUF_SIZE )
    {
      if ( !quiet )
        error ( "mms_recv_cmd_packet", "incoming packet is too big for me" );

      return -1;
    }

  if ( mms_recv_packet ( s, &tmp, *packet_len, quiet ) != 0 )
    {
      if ( !quiet )
        error ( "mms_recv_cmd_packet", "unable to get packet body" );

      return -1;
    }

  memcpy ( &pak->buf[12], tmp.buf, *packet_len );

  return ( mms_get_32(pak->buf, 36) & 0xFFFF );
}


/*
 * Gets the header of the ASF/WMV stream/file.
 * Returns the amount of bytes received, -1 if we encounter an error.
 */
static ssize_t
mms_recv_header_packet ( MMS *mms,
                         MMS_PACKET *pak )
{
  MMS_PACKET pre_header;
  MMS_PACKET tmp;
  ssize_t header_len;
  ssize_t packet_len;

#ifdef ENABLE_DEBUG
  ssize_t i;

  fprintf ( stdout, "\n-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
  fprintf ( stdout, " getting header from server\n\n" );
#endif

  header_len = 0;

  while ( 1 )
    {
      if ( mms_recv_packet ( mms->socket, &pre_header, 8, mms->quiet ) != 0 )
        {
          if ( !mms->quiet )
            error ( "mms_recv_header_packet", "unable to get pre-header" );

          return -1;
        }

#ifdef ENABLE_DEBUG
      for ( i=0; i<8; ++i )
        fprintf ( stdout, "pre_header.buf[%d] = 0x%02x (%d)\n", i, pre_header.buf[i], pre_header.buf[i] );
#endif
    
      if ( pre_header.buf[4] == 0x02 )
        {
          packet_len = ( pre_header.buf[7] << 8 | pre_header.buf[6] ) - 8;

#ifdef ENABLE_DEBUG
          fprintf ( stdout, "\nasf header packet detected, len=%d\n", packet_len );
#endif

          if ( mms_recv_packet ( mms->socket, &tmp, packet_len, mms->quiet ) != 0 )
            {
              if ( !mms->quiet )
                error ( "mms_recv_header_packet", "unable to get header" );

              return -1;
            }

          if ( ( header_len + packet_len ) > MMS_BUF_SIZE )
            {
              if ( !mms->quiet )
                error ( "mms_recv_header_packet", "total header length is too big for me" );

              return -1;
            }

          memcpy ( &pak->buf[header_len], tmp.buf, packet_len );

          header_len += packet_len;

          /* test if end of header is reached */
          if ( ( pak->buf[header_len-1] == 1) &&
               ( pak->buf[header_len-2] == 1 ) )
            {
#ifdef ENABLE_DEBUG
              fprintf ( stdout, "\n\n" );
#endif
              return header_len;
            }
        }
      else /* we might receive an ack or an error */
        {
          int command = mms_recv_cmd_packet ( mms->socket, &tmp, &packet_len, 8, mms->quiet );

          if ( command == -1 )
            {
              if ( !mms->quiet )
                error ( "mms_recv_header_packet", "unable to get cmd packet" );

              return -1;
            }

          memcpy ( tmp.buf, pre_header.buf, 8 );

#ifdef ENABLE_DEBUG
          mms_print_packet ( &tmp, packet_len, MMS_SERVER );
#endif
          if ( command == MMS_CMD_PING )
            mms_send_packet ( mms, MMS_CMD_PONG, 0, 0, 0, tmp.buf );
          else if ( command != MMS_CMD_HEADER_DATA )
            {
              if ( !mms->quiet )
                error ( "mms_recv_header_packet", "unknown command 0x%02x\n", command );

              return -1;
            }
        }

#ifdef ENABLE_DEBUG
      fprintf ( stdout, "\n\n" );
#endif
    }
}


/*
 * Interprates the header (media packet size, etc.).
 * Returns the size of a media packet (should be != 0 ).
 */
static ssize_t
mms_interp_header ( MMS *mms,
                    const uint8_t *header,
                    const ssize_t header_len )
{
  int i;
  ssize_t packet_length = 0;

  mms->expected_file_size = header_len;

#ifdef ENABLE_DEBUG
  fprintf ( stdout, "\n-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
  fprintf ( stdout, " header interpretation\n\n" );
#endif

  /* parse header */
  mms->num_stream_ids = 0;
  i = 30;

  while ( i < header_len )
    {
      uint64_t  guid_1, guid_2, length;

      guid_2 = ((uint64_t) header[i])       | ((uint64_t) header[i+1]<<8)  | ((uint64_t) header[i+2]<<16) | ((uint64_t) header[i+3]<<24) |
               ((uint64_t) header[i+4]<<32) | ((uint64_t) header[i+5]<<40) | ((uint64_t) header[i+6]<<48) | ((uint64_t) header[i+7]<<56);

      i += 8;

      guid_1 = ((uint64_t)header[i])       | ((uint64_t)header[i+1]<<8)  | ((uint64_t)header[i+2]<<16) | ((uint64_t)header[i+3]<<24) |
               ((uint64_t)header[i+4]<<32) | ((uint64_t)header[i+5]<<40) | ((uint64_t)header[i+6]<<48) | ((uint64_t)header[i+7]<<56);

      i += 8;

#ifdef ENABLE_DEBUG
      fprintf ( stdout, "guid found: %016llx%016llx\n", guid_1, guid_2 );
#endif

      length = ((uint64_t)header[i])       | ((uint64_t)header[i+1]<<8)  | ((uint64_t)header[i+2]<<16) | ((uint64_t)header[i+3]<<24) |
               ((uint64_t)header[i+4]<<32) | ((uint64_t)header[i+5]<<40) | ((uint64_t)header[i+6]<<48) | ((uint64_t)header[i+7]<<56);

      i += 8;

      if ( ( guid_1 == 0x6553200cc000e48eULL ) && ( guid_2 == 0x11cfa9478cabdca1ULL ) )
        {
          packet_length = mms_get_32 ( header, i+92-24 );
#ifdef ENABLE_DEBUG
          fprintf ( stdout, "file object, packet length=%d (%d)\n", packet_length, mms_get_32(header, i+96-24) );
#endif
        }
      else if  ( ( guid_1 == 0x6553200cc000e68eULL ) && ( guid_2 == 0x11cfa9b7b7dc0791ULL ) )
        {
          int stream_id = header[i+48] | header[i+49] << 8;

          if ( mms->num_stream_ids < 20 )
            {
              mms->stream_ids[mms->num_stream_ids] = stream_id;
              ++mms->num_stream_ids;
            }
          else
            {
              if ( !mms->quiet )
                warning ( "mms_interp_header", "too many streams, stream \'%d\' skipped", stream_id );
            }

#ifdef ENABLE_DEBUG
          fprintf ( stdout, "stream object, stream id: %d\n", stream_id );
#endif
        }
      else if ( ( guid_1 == 0x6cce6200aa00d9a6ULL ) && ( guid_2 == 0x11cf668e75b22636ULL ) )
        {
          mms->expected_file_size += length;
#ifdef ENABLE_DEBUG
          fprintf ( stdout, "data object\n" );
#endif
        }
#ifdef ENABLE_DEBUG
      else if ( ( guid_1 == 0x6cce6200aa00d9a6ULL ) && ( guid_2 == 0x11cf668e75b22630ULL ) )
        fprintf ( stdout, "header object\n" );
      else
        fprintf ( stdout, "unknown object\n" );

      fprintf ( stdout, "length: 0x%x\n", length );
#endif

      i += length-24;
  }

  return packet_length;
}


/*
 * Gets a media packet.
 * Returns the amount of bytes in the media packet, 0 if EOF, -2 if ACK, -1 either.
 */
static ssize_t
mms_recv_media_packet ( MMS *mms,
                        MMS_PACKET *pak )
{
  MMS_PACKET pre_header;
  ssize_t packet_len;

#ifdef ENABLE_DEBUG
  ssize_t i;

  fprintf ( stdout, "\n-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
  fprintf ( stdout, " getting media packet from server\n\n" );
#endif

  if ( mms_recv_packet ( mms->socket, &pre_header, 8, mms->quiet ) != 0 )
    {
      if ( !mms->quiet )
        error ( "mms_recv_media_packet", "unable to get pre-header" );

      return -1;
    }

#ifdef ENABLE_DEBUG
  for ( i=0; i<8; i++ )
    fprintf ( stdout, "pre_header[%d] = 0x%02x (%d)\n", i, pre_header.buf[i], pre_header.buf[i] );
#endif

  if ( pre_header.buf[4] == 0x04 )
    {
      packet_len = (pre_header.buf[7] << 8 | pre_header.buf[6]) - 8;

#ifdef ENABLE_DEBUG
      fprintf ( stdout, "asf media packet detected, len=%d\n", packet_len );
#endif

      memset ( pak->buf, 0, mms->media_packet_len );

      if ( mms_recv_packet (mms->socket, pak, packet_len, mms->quiet) != 0 )
        {
          if ( !mms->quiet )
            error ( "mms_recv_media_packet", "unable to get media packet" );

          return -1;
        }
    }
  else
    {
      int command = mms_recv_cmd_packet ( mms->socket, pak, &packet_len, 8, mms->quiet );

      if ( command == -1 )
        {
          if ( !mms->quiet )
            error ( "mms_recv_media_packet", "unable to get cmd packet" );

          return -1;
        }

      memcpy ( pak->buf, pre_header.buf, 8 );

#ifdef ENABLE_DEBUG
      mms_print_packet ( pak, packet_len, MMS_SERVER );
#endif

      if ( command == MMS_CMD_PING )
        {
          mms_send_packet ( mms, MMS_CMD_PONG, 0, 0, 0, pak->buf );
          return -2;
        }
      else if ( command == MMS_CMD_END_OF_STREAM )
        return 0;
      else if ( command == MMS_CMD_STREAM_SELECT_ACK ) /* it happens sometimes */
        return -2;
      else if ( command == MMS_CMD_READY_TO_STREAM ) /* it always happen before the first media packet */
        return -2;
      else
        {
          if ( !mms->quiet )
            error ( "mms_recv_media_packet", "unknown command 0x%02x\n", command );

          return -1;
        }
    }

#ifdef ENABLE_DEBUG
  fprintf ( stdout, "\n\n" );
#endif

  return packet_len;
}


/*
 * Creates An MMS Struct
 */
MMS *
mms_create ( const char *url,
             FILE *out,
             const int quiet )
{
  MMS *mms;
  register const char *sep = strchr ( &url[6], '/' );
  register const int url_len = strlen ( url );

  if ( sep == NULL )
    {
      if ( !quiet )
        error ( "mms_create", "url seems to be malformed" );

      return NULL;
    }

  if ( ( mms = ( MMS * ) malloc ( sizeof(MMS) ) ) == NULL )
    {
      if ( !quiet )
        error ( "mms_create", "unable to allocate memory" );

      return NULL;
    }

  mms->host = malloc ( ( sep - &url[6] ) + 1 );
  strncpy ( mms->host, &url[6], (sep - &url[6]) );
  mms->host[sep-&url[6]] = '\0'; /* strndup */

  mms->path = strdup ( sep+1 );
  mms->out = out;
  mms->seq_num = 0;
  mms->expected_file_size = 0;
  mms->is_live = MMS_NO_LIVE;

  /* we try to guess stream type with the filename extension */
  if ( ( url[url_len-4] == '.' ) &&
       ( url[url_len-3] == 'w' ) &&
       ( url[url_len-2] == 'm' ) &&
       ( url[url_len-1] == 'v' ) )
    mms->stream_type = MMS_WMV;
  else
    mms->stream_type = MMS_ASF;

  mms->quiet = quiet;

#ifdef ENABLE_DEBUG
  fprintf ( stdout, "host is %s\n", mms->host );
  fprintf ( stdout, "path is %s\n", mms->path );
#endif

  return mms;
}


/*
 * Establishes a connection with the MMS server.
 * Returns 0 if success, -1 either.
 */
int
mms_connect ( MMS* mms )
{
  struct sockaddr_in   sa;
  struct hostent      *hp;

  if ( mms == NULL )
    return -1;

  if ( ( hp = gethostbyname(mms->host) ) == NULL )
    {
      if ( !mms->quiet )
        error ( "mms_connect", "unable to resolve host name" );

      return -1;
    }

  bcopy ( (char *) hp->h_addr, (char *) &sa.sin_addr, hp->h_length );
  sa.sin_family = hp->h_addrtype;
  sa.sin_port = htons ( 1755 );

  if ( ( mms->socket = socket(hp->h_addrtype, SOCK_STREAM, 0) ) == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_connect", "socket() said: %s", strerror(errno) );

      return  -1;
    }

  if ( connect ( mms->socket, (struct sockaddr *) &sa, sizeof sa ) != 0 )
    {
      if ( !mms->quiet )
        error ( "mms_connect", "connect() said: %s", strerror(errno) );

      return -1;
    }

  return 0;
}


/*
 * Handshake with the MMS server.
 * ( we don't care about server's answers )
 */
int
mms_handshake ( MMS *mms )
{
  MMS_PACKET pak;
  int cmd = MMS_CMD_PING;
  ssize_t packet_len;
  char str [ 1024 ];
  uint8_t data [ 2092 ];

  if ( mms == NULL )
    return -1;

  /* who we are */
  memset ( data, 0, sizeof (data) );
  snprintf ( str, sizeof(str), "\034\003NSPlayer/7.0.0.1956; {33715801-BAB3-9D85-24E9-03B90328270A}; Host: %s", mms->host );
  mms_string_utf16 ( data, (unsigned char *) str, strlen(str)+2 );
  mms_send_packet ( mms, MMS_CMD_HELLO, 0, 0x0004000b, strlen(str) * 2+8, data );

  cmd = MMS_CMD_PING;
  while ( cmd == MMS_CMD_PING )
    {
      cmd = mms_recv_cmd_packet ( mms->socket, &pak, &packet_len, 0, mms->quiet );

      if ( cmd == MMS_CMD_PING )
        mms_send_packet ( mms, MMS_CMD_PONG, 0, 0, 0, pak.buf );
    }

  if ( cmd == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_handshake", "unable to get cmd packet" );

      return -1;
    }

#ifdef ENABLE_DEBUG
    mms_print_packet ( &pak, packet_len, MMS_SERVER );
#endif

  /* transport protocol selection */
  memset ( &data, 0, sizeof(data) );
  mms_string_utf16 ( &data[8], (unsigned char *) "\002\000\\\\192.168.0.129\\TCP\\1037\0000", 28 );
  mms_send_packet ( mms, MMS_CMD_PROTOCOL_SELECT, 0, 0, 28*2+8, data );

  cmd = MMS_CMD_PING;
  while ( cmd == MMS_CMD_PING )
    {
      cmd = mms_recv_cmd_packet ( mms->socket, &pak, &packet_len, 0, mms->quiet );

      if ( cmd == MMS_CMD_PING )
        mms_send_packet ( mms, MMS_CMD_PONG, 0, 0, 0, pak.buf );
    }

  if ( cmd == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_handshake", "unable to get cmd packet" );

      return -1;
    }

#ifdef ENABLE_DEBUG
    mms_print_packet ( &pak, packet_len, MMS_SERVER );
#endif

  /* requested file */
  mms_string_utf16 ( &data[8], (unsigned char *) mms->path, strlen(mms->path) );
  memset ( data, 0, 8 );
  mms_send_packet ( mms, MMS_CMD_FILE_REQUEST, 0, 0, strlen(mms->path)*2+12, data );

  cmd = MMS_CMD_PING;
  while ( cmd == MMS_CMD_PING )
    {
      cmd = mms_recv_cmd_packet ( mms->socket, &pak, &packet_len, 0, mms->quiet );

      if ( cmd == MMS_CMD_PING )
        mms_send_packet ( mms, MMS_CMD_PONG, 0, 0, 0, pak.buf );
    }

  if ( cmd == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_handshake", "unable to get cmd packet" );

      return -1;
    }

  if ( cmd == MMS_CMD_STREAM_INFOS )
    {
      if ( mms_get_32 ( pak.buf, 48 ) == -1 )
        {
          if ( !mms->quiet )
            error ( "mms_handshake", "stream infos are not available" );

          return -1;
        }

      if ( ( mms_get_32 ( pak.buf, 72 ) == 0 ) || ( mms_get_32 ( pak.buf, 72 ) == 0xFFFFFFFF ) )
        {
          mms->is_live = MMS_LIVE;

          if ( !mms->quiet )
            warning ( NULL, "stream seems to be live, an error may occur" );
        }
    }

#ifdef ENABLE_DEBUG
    mms_print_packet ( &pak, packet_len, MMS_SERVER );
#endif

  return 0;
}


/*
 * Gets ASF/WMV stream/file header
 */
ssize_t
mms_write_stream_header ( MMS *mms )
{
  MMS_PACKET pak;
  ssize_t len;

  if ( mms == NULL )
    return -1;

  /* give me the stream header please... */
  memset ( pak.buf, 0, 40 );
  pak.buf[32] = 2;
  mms_send_packet ( mms, MMS_CMD_HEADER_REQUEST, 1, 0, 40, pak.buf );

  len = mms_recv_header_packet ( mms, &pak );

  if ( len == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_get_stream_header", "unable to get stream header\n" );

      return -1;
    }
  else if ( len == 0 )
    {
      if ( !mms->quiet )
        error ( "mms_get_stream_header", "no stream header available, file may not exist" );

      return -1;
    }

  fwrite ( pak.buf, len, 1, mms->out );

  mms->media_packet_len = mms_interp_header ( mms, pak.buf, len );

  if ( mms->media_packet_len > MMS_BUF_SIZE )
    {
      if ( !mms->quiet )
        error ( "mms_get_stream_header", "media packet length is too big for me" );

      return -1;
    }

  return len;
}


/*
 * Tells the server we're ready to rip the stream from the beginning (don't know what happens if it's a live)
 * Returns 0 if success, -1 either.
 */
int
mms_begin_rip ( MMS *mms )
{
  int i;
  uint8_t data [ 1024 ];
  MMS_PACKET pak;
  ssize_t packet_len;
  int cmd;

  if ( mms == NULL )
    return -1;

  /* media stream selection */
  memset ( data, 0, 40 );

  for ( i=1; i<mms->num_stream_ids; ++i )
    {
      data [ (i-1) * 6 + 2 ] = 0xFF;
      data [ (i-1) * 6 + 3 ] = 0xFF;
      data [ (i-1) * 6 + 4 ] = mms->stream_ids[i];
      data [ (i-1) * 6 + 5 ] = 0x00;
    }

  /* we add the final bytes of the stream selector, necessary for ASF streams only */
  if ( mms->stream_type == MMS_ASF )
    {
      data [ (mms->num_stream_ids-1) * 6 + 0 ] = 0x00;
      data [ (mms->num_stream_ids-1) * 6 + 1 ] = 0x00;
      data [ (mms->num_stream_ids-1) * 6 + 2 ] = 0x00;
      data [ (mms->num_stream_ids-1) * 6 + 3 ] = 0x20;
      data [ (mms->num_stream_ids-1) * 6 + 4 ] = 0xac;
      data [ (mms->num_stream_ids-1) * 6 + 5 ] = 0x40;
      data [ (mms->num_stream_ids-1) * 6 + 6 ] = 0x02;

      mms_send_packet ( mms, MMS_CMD_STREAM_SELECT, mms->num_stream_ids, 0xFFFF | mms->stream_ids[0] << 16, (mms->num_stream_ids-1)*6+2+4, data );
    }
  else
    {
      mms_send_packet ( mms, MMS_CMD_STREAM_SELECT, mms->num_stream_ids, 0xFFFF | mms->stream_ids[0] << 16, (mms->num_stream_ids-1)*6+2, data );
    }

  cmd = MMS_CMD_PING;
  while ( cmd == MMS_CMD_PING )
    {
      cmd = mms_recv_cmd_packet ( mms->socket, &pak, &packet_len, 0, mms->quiet );

      if ( cmd == MMS_CMD_PING )
        mms_send_packet ( mms, MMS_CMD_PONG, 0, 0, 0, pak.buf );
    }

  if ( cmd == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_begin_rip", "unable to get server\'s confirmation" );

      return -1;
    }

#ifdef ENABLE_DEBUG
  mms_print_packet ( &pak, packet_len, MMS_SERVER );
#endif


  /* ready... */
  memset ( data, 0, 40 );

  for ( i=8; i<16; ++i )
    data[i] = 0xFF;

  data[20] = 0x04;

  mms_send_packet ( mms, MMS_CMD_START_PACKET, 1, 0xFFFF | mms->stream_ids[0] << 16, 24, data );

  return 0;
}


/*
 * Writes a media packet.
 * Returns the amount of bytes written, 0 if EOF, -1 either.
 */
ssize_t
mms_write_stream_data ( MMS *mms )
{
  ssize_t len;
  MMS_PACKET pak;

  if ( mms == NULL )
    return 0;
  
  do
    {
      len = mms_recv_media_packet ( mms, &pak );
    }
  while ( len == -2 ); /* while we receive ACK packets */

  if ( len == 0 )
    return 0;
  else if ( len == -1 )
    {
      if ( !mms->quiet )
        error ( "mms_write_stream_data", "mms_recv_media_packet failed" );

      return -1;
    }

  fwrite ( pak.buf, mms->media_packet_len, 1, mms->out ); /* we might have read less than the arbitrary media packet size */

  return mms->media_packet_len;
}


/*
 * Closes the connection.
 */
void
mms_disconnect ( MMS *mms )
{
  uint8_t data [ 1024 ];

  if ( mms == NULL )
    return;

  mms_send_packet ( mms, MMS_CMD_BYE_BYE, 0, 0, 0, data );

  if ( ( shutdown(mms->socket, SHUT_RDWR) | close(mms->socket)) != 0 )
    if ( !mms->quiet )
      warning ( "mms_disconnect", "unable to close the socket properly" );

  return;
}


/*
 * Destroys An MMS Struct
 */
void
mms_destroy ( MMS *mms )
{
  if ( mms == NULL )
    return;

  if ( mms->host != NULL )
    free ( mms->host );

  if ( mms->path != NULL )
    free ( mms->path );

  free ( mms );

  return;
}
